#[derive(Debug,PartialEq,Eq,Clone)]
pub struct Token<'i> {
    pub kind: TokenKind<'i>,
    pub offs: usize,
}

#[derive(Debug,PartialEq,Eq,Clone)]
pub enum TokenKind<'i> {
    Tab,
    Space,
    Newline,
    Word(&'i str),
}

impl<'i> TokenKind<'i> {
    pub fn len(&self) -> usize {
        match self {
            TokenKind::Tab => 1,
            TokenKind::Space => 1,
            TokenKind::Newline => 1,
            TokenKind::Word(w) => w.len(),
        }
    }

    pub fn explain(&self) -> String {
        match self {
            TokenKind::Tab => "tab".to_string(),
            TokenKind::Space => "space".to_string(),
            TokenKind::Newline => "newline".to_string(),
            TokenKind::Word(w) => format!("literal '{}'", w),
        }
    }
}

pub struct Tokens<'i> {
    pub tokens: Vec<Token<'i>>,
    pub underlying: &'i str,
}

impl<'i> Tokens<'i> {
    pub fn parse(s: &'i str) -> Self {
        let mut res: Vec<Token<'i>> = vec![];
        let mut cur: Option<usize> = None;

        for (i, ch) in s.char_indices() {
            if let Some(c) = cur {
                // Continue word.
                let n = match ch {
                    '\t' => Some(Token {
                        kind: TokenKind::Tab,
                        offs: i,
                    }),
                    ' ' => Some(Token {
                        kind: TokenKind::Space,
                        offs: i,
                    }),
                    '\n' => Some(Token {
                        kind: TokenKind::Newline,
                        offs: i,
                    }),
                    _ => None,
                };
                if let Some(t) = n {
                    res.push(Token {
                        kind: TokenKind::Word(&s[c..i]),
                        offs: c
                    });
                    res.push(t);
                    cur = None;
                }
            } else {
                // Begin new token if necessary.
                match ch {
                    '\t' => res.push(Token {
                        kind: TokenKind::Tab,
                        offs: i,
                    }),
                    ' ' => res.push(Token {
                        kind: TokenKind::Space,
                        offs: i,
                    }),
                    '\n' => res.push(Token {
                        kind: TokenKind::Newline,
                        offs: i,
                    }),
                    _ => cur = Some(i),
                }
            }
        }
        match cur {
            None => (),
            Some(c) => res.push(Token {
                kind: TokenKind::Word(&s[c..]),
                offs: c,
            }),
        }

        Self {
            tokens: res,
            underlying: s,
        }
    }

    pub fn view<'p>(&'p self) -> TokensView<'i, 'p> {
        TokensView {
            parent: &self,
            start: 0,
        }
    }
}

pub struct TokensView<'i, 'p> {
    pub parent: &'p Tokens<'i>,
    pub start: usize,
}

impl<'i, 'p> TokensView<'i, 'p> {
    pub fn iter<'b>(&'b self) -> TokenIterator<'i, 'b>
    where
        'p: 'b
    {
        TokenIterator {
            cur: self.start,
            end: self.parent.tokens.len(),
            underlying: &self.parent.tokens,
        }
    }

    pub fn len(&self) -> usize {
        return self.parent.tokens.len() - self.start
    }

    pub fn sub(&self, split: usize) -> TokensView<'i, 'p> {
        TokensView {
            parent: &self.parent,
            start: split + self.start,
        }
    }
}

pub struct TokenIterator<'i, 'p>
where
    'i: 'p
{
    cur: usize,
    end: usize,
    underlying: &'p Vec<Token<'i>>,
}

impl<'i, 'p> Iterator for TokenIterator<'i, 'p>
{
    type Item = &'p Token<'i>;
    fn next(&mut self) -> Option<&'p Token<'i>> {
        if self.cur >= self.end {
            return None;
        }
        let res = Some(&self.underlying[self.cur]);
        self.cur += 1;
        res
    }
}

