use std::io::Write;

pub mod lex;
pub mod parse;

#[cfg(test)]
mod tests {
    use super::lex::{
        TokenKind, Tokens
    };
    use super::parse::{
        Pattern, Captured,
        Stanza, EOL,
    };
    const HEADER: &'static str = "GMA
*GABRIEL_ASCIIEXPORT\t3
*MODEL_TYPE\tBasic Model
*SCENE
{
\t*SCENE_FILENAME\tfoo.max
}
";
    #[test]
    fn test_lex() {
        let parsed = Tokens::parse(HEADER);
        let kinds: Vec<TokenKind> = parsed.tokens.into_iter().map(|t| t.kind).collect();
        assert_eq!(kinds, vec![
            TokenKind::Word("GMA"), TokenKind::Newline,
            TokenKind::Word("*GABRIEL_ASCIIEXPORT"), TokenKind::Tab, TokenKind::Word("3"), TokenKind::Newline,
            TokenKind::Word("*MODEL_TYPE"), TokenKind::Tab, TokenKind::Word("Basic"), TokenKind::Space, TokenKind::Word("Model"), TokenKind::Newline,
            TokenKind::Word("*SCENE"), TokenKind::Newline,
            TokenKind::Word("{"), TokenKind::Newline,
            TokenKind::Tab, TokenKind::Word("*SCENE_FILENAME"), TokenKind::Tab, TokenKind::Word("foo.max"), TokenKind::Newline,
            TokenKind::Word("}"), TokenKind::Newline,
        ]); 
    }

    #[test]
    fn test_parse_simple() {
        let example = Tokens::parse(HEADER);
        let view = example.view();
        let res = match vec![
            Box::new(TokenKind::Word("GMA")) as Box<dyn Pattern>,
            Box::new(TokenKind::Newline),

            Box::new(Stanza("*GABRIEL_ASCIIEXPORT", (TokenKind::Word("3"),))),
            Box::new(Stanza("*MODEL_TYPE", (EOL,))),

            // TODO blocks
            Box::new(TokenKind::Word("*SCENE")), Box::new(TokenKind::Newline),
            Box::new(TokenKind::Word("{")), Box::new(TokenKind::Newline),

            Box::new(Stanza("*SCENE_FILENAME", (EOL,))),
            Box::new(TokenKind::Word("}")), Box::new(TokenKind::Newline),
        ].matches(&view) {
            Ok(v) => v,
            Err(e) => {
                e.explain(&example, "test_txt.gmf");
                panic!("parse failed");
            }
        };

        let rest = view.sub(res.consumed);
        assert_eq!(rest.len(), 0);
        println!("{:#?}", res.captured);
        let v = match res.captured {
            Captured::Vec(v) => v,
            _ => panic!("expected top-level vec"),
        };
        assert_eq!(v.len(), 11);
        match &v[2] {
            Captured::Stanza { key, .. } => {
                assert_eq!(*key, "*GABRIEL_ASCIIEXPORT");
            },
            _ => panic!("expected stanza in 2nd element"),
        }
        match &v[3] {
            Captured::Stanza { key, values, .. } => {
                assert_eq!(*key, "*MODEL_TYPE");
                assert_eq!(values.as_ref(), &Captured::Tuple1(Box::new(Captured::Str("Basic Model"))));
            },
            _ => panic!("expected stanza in 3rd element"),
        }
    }
}

#[derive(Debug, Clone)]
pub struct WriteError {
    pub msg: String,
}

pub type WriteResult<T> = Result<T, WriteError>;

pub trait Serializable: Sized {
    fn write<W: std::io::Write, S: std::string::ToString>(&self, name: S, r: &mut WriteStream<W>) -> WriteResult<()>;
}

pub struct WriteStream<W: std::io::Write> {
    backing: std::io::BufWriter<W>,
    depth: usize,
}

pub trait Atom {
    fn atom(&self) -> String;
}

impl Atom for str {
    fn atom(&self) -> String {
        self.to_string()
    }
}

impl Atom for String {
    fn atom(&self) -> String {
        self.to_string()
    }
}

impl Atom for f32 {
    fn atom(&self) -> String {
        format!("{:.6}", self)
    }
}

impl<const N: usize> Serializable for [f32; N] {
    fn write<W: std::io::Write, S: std::string::ToString>(&self, name: S, w: &mut WriteStream<W>) -> WriteResult<()> {
        let fields = self.iter().map(|v| format!("{:.6}", v)).collect::<Vec<String>>().join("\t");
        w.emit(&format!("{} {}", name.to_string(), fields))
    }
}

impl Atom for i32 {
    fn atom(&self) -> String {
        format!("{}", self)
    }
}

impl Atom for u32 {
    fn atom(&self) -> String {
        format!("{}", self)
    }
}

impl Atom for u8 {
    fn atom(&self) -> String {
        format!("{}", self)
    }
}

impl Atom for usize {
    fn atom(&self) -> String {
        format!("{}", self)
    }
}

impl<I: Serializable> Serializable for Option<I> {
    fn write<W: std::io::Write, S: std::string::ToString>(&self, name: S, w: &mut WriteStream<W>) -> WriteResult<()> {
        if let Some(s) = self {
            s.write(name, w)?;
        }
        Ok(())
    }
}

impl<I: Serializable> Serializable for Vec<I> {
    fn write<W: std::io::Write, S: std::string::ToString>(&self, name: S, w: &mut WriteStream<W>) -> WriteResult<()> {
        if name.to_string() == "*" {
            w.emit_pair("*COUNT", &self.len())?;
        } else {
            w.emit_pair(&(name.to_string() + "_COUNT"), &self.len())?;
        }
        for elem in self.iter() {
            elem.write(name.to_string(), w)?;
        }
        Ok(())
    }
}

impl<A: Atom> Serializable for A {
    fn write<W: std::io::Write, S: std::string::ToString>(&self, name: S, w: &mut WriteStream<W>) -> WriteResult<()> {
        if self.atom().len() > 0 {
            w.emit_pair(&name.to_string(), self)?;
        }
        Ok(())
    }
}

impl <W: std::io::Write> WriteStream<W> {
    pub fn new(w: W) -> Self {
        Self {
            backing: std::io::BufWriter::new(w),
            depth: 0,
        }
    }

    pub fn error<S: std::string::ToString>(&self, msg: S) -> WriteError {
        WriteError {
            msg: msg.to_string(),
        }
    }

    pub fn emit<S: Atom + ?Sized>(&mut self, elem: &S) -> WriteResult<()> {
        match write!(self.backing, "{}{}\n", "\t".repeat(self.depth), elem.atom()) {
            Ok(_) => return Ok(()),
            Err(_) => return Err(self.error("write failed")),
        }
    }

    pub fn emit_pair<S: Atom + ?Sized, T: Atom + ?Sized>(&mut self, a: &S, b: &T) -> WriteResult<()> {
        let a = a.atom();
        if a == "" || a == "*" {
            return Ok(());
        }
        let b = b.atom();
        match write!(self.backing, "{}{}\t{}\n", "\t".repeat(self.depth), a, b) {
            Ok(_) => return Ok(()),
            Err(_) => return Err(self.error("write failed")),
        }
    }

    pub fn push<S: std::string::ToString>(&mut self, el: S) -> WriteResult<()> {
        self.emit(&el.to_string())?;
        self.emit("{")?;
        self.depth += 1;
        Ok(())
    }

    pub fn pop(&mut self) -> WriteResult<()> {
        if self.depth == 0 {
            return Err(self.error("stack underflow"));
        }
        self.depth -= 1;
        self.emit("}")?;
        Ok(())
    }
}
