use crate::gma::lex::{
    Token, TokenKind, Tokens, TokensView,
};

#[derive(Debug)]
pub struct Match<'i> {
    pub consumed: usize,
    pub captured: Captured<'i>,
}

pub type MatchResult<'i> = Result<Match<'i>, MatchError>;

#[derive(Debug,PartialEq,Eq)]
pub enum Captured<'i> {
    // low level
    Void,
    Vec(Vec<Captured<'i>>),
    Tuple1(Box<Captured<'i>>),
    Tuple2(Box<Captured<'i>>, Box<Captured<'i>>),
    Tuple3(Box<Captured<'i>>, Box<Captured<'i>>, Box<Captured<'i>>),
    Tuple4(Box<Captured<'i>>, Box<Captured<'i>>, Box<Captured<'i>>, Box<Captured<'i>>),
    Token(Token<'i>),

    // mid level
    U32(u32),

    // high level
    Str(&'i str),
    Stanza {
        indent: Vec<Captured<'i>>,
        key: &'i str,
        values: Box<Captured<'i>>,
    }
}

#[derive(Debug)]
pub enum MatchError {
    PatternMismatch {
        token_offset: usize,
        wanted: Box<dyn Pattern>,
    },
    EOF {
        token_offset: usize,
    },
}

impl MatchError {
    pub fn explain<'i>(&self, tokens: &Tokens<'i>, filename: &str) {
        let offset = *match self {
            MatchError::PatternMismatch { token_offset, ..} => token_offset,
            MatchError::EOF { token_offset } => token_offset,
        };
        let token = &tokens.tokens[offset];
        let byte_offset = token.offs;

        let lines: Vec<&str> = tokens.underlying.split("\n").collect();

        let mut rolling = 0usize;
        let mut line_no: Option<usize> = None;
        let mut byte_in_line: Option<usize> = None;
        for (i, line) in lines.iter().enumerate() {
            let start = rolling;
            let end = rolling + line.bytes().len();
            if start <= byte_offset && byte_offset <= end {
                line_no = Some(i);
                byte_in_line = Some(byte_offset - start);
                break;
            }
            // TODO: fix assumption of no consecutive newlines?
            rolling = end + 1;
        }
        let line_no = line_no.unwrap();
        let byte_in_line = byte_in_line.unwrap();

        // Replace all tabs with four spaces when visualizing the line, and add to byte_in_line to
        // compensate.
        let line = lines[line_no];
        let visual = byte_in_line + lines[line_no][..byte_in_line].matches("\t").count() * 3;
        let line = line.replace("\t", "    ");

        let msg = match self {
            MatchError::PatternMismatch { wanted, .. } => format!("unexpected {}, wanted {}", token.kind.explain(), wanted.explain()),
            MatchError::EOF{ .. } => format!("unexpected end of flie"),
        };
        println!("error: {}", msg);
        println!("      --> {}:{}:{}", filename, line_no+1, byte_in_line);
        println!("       |");
        println!(" {:>5} | {}", line_no+1, line);
        println!("       | {}{}",
                 " ".repeat(visual),
                 "^".repeat(token.kind.len()));
    }
}

struct TokenReader<'i, 'p> {
    view: TokensView<'i, 'p>,
    consumed: usize,
}

impl<'i, 'p> TokenReader<'i, 'p> {
    fn new(view: &TokensView<'i, 'p>) -> Self {
        Self {
            view: view.sub(0),
            consumed: 0,
        }
    }

    fn expect<P: Pattern + ?Sized>(&mut self, pattern: &P) -> Result<Captured<'i>, MatchError> {
        let m = pattern.matches(&self.view)?;
        self.view = self.view.sub(m.consumed);
        self.consumed += m.consumed;
        Ok(m.captured)
    }

    fn done(self, captured: Captured<'i>) -> MatchResult {
        Ok(Match {
            consumed: self.consumed,
            captured, 
        })
    }
}

pub trait Pattern: core::fmt::Debug {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i>;
    fn explain(&self) -> String;
}

#[derive(Debug,Clone)]
struct Whitespace;

impl Pattern for Whitespace {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        // Not using TokenReader here as we don't have alternatives support yet.
        let mut res: Vec<Captured<'i>> = vec![];
        for t in l.iter() {
            match t.kind {
                TokenKind::Tab | TokenKind::Space => {
                    res.push(Captured::Token(t.clone()))
                },
                _ => break,
            }
        }
        if res.len() > 0 {
            Ok(Match {
                consumed: res.len(),
                captured: Captured::Vec(res),
            })
        } else {
            Err(MatchError::PatternMismatch {
                token_offset: l.start,
                wanted: Box::new(self.clone()),
            })
        }
    }

    fn explain(&self) -> String {
        "space or tab".to_string()
    }
}

impl Pattern for TokenKind<'static> {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        // Not using TokenReader as this is a terminal.
        if l.len() < 1 {
            return Err(MatchError::EOF {
                token_offset: l.start,
            })
        }
        let v = l.iter().next().unwrap();
        if v.kind != *self {
            return Err(MatchError::PatternMismatch {
                token_offset: l.start,
                wanted: Box::new(self.clone()),
            })
        }
        return Ok(Match {
            consumed: 1,
            captured: Captured::Token(v.clone()),
        })
    }

    fn explain(&self) -> String {
        self.explain()
    }
}

impl Pattern for Vec<Box<dyn Pattern>> {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        let mut res: Vec<Captured<'i>> = vec![];
        let mut r = TokenReader::new(l);
        for t in self.iter() {
            res.push(r.expect(t.as_ref())?);
        }
        r.done(Captured::Vec(res))
    }
    fn explain(&self) -> String {
        format!("sequence [{}]", self
                .iter().map(|el| el.explain())
                .collect::<Vec<String>>().join(","))
    }
}

impl Pattern for () {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        let r = TokenReader::new(l);
        r.done(Captured::Void)
    }
    fn explain(&self) -> String {
        "nothing".to_string()
    }
}

macro_rules! mktuple {
    ( $cons:ident, $count:expr, [ $($ty:ident),+ ], [ $($name:ident),+ ] ) => {
        impl <$($ty),*> Pattern for ( $($ty),+ , )
            where $( $ty : Pattern ),+
        {
            fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
                let mut r = TokenReader::new(l);
                let mut i = 0;
                let ( $( $name ),+ , ) = self;
                $(
                    let $name = r.expect($name)?;
                    i += 1;
                    if i != $count {
                        r.expect(&Whitespace)?;
                    }
                )+
                r.done(Captured::$cons(
                    $( 
                        Box::new($name)
                    ),+
                ))
            }
            fn explain(&self) -> String {
                let ( $( $name ),+ , ) = self;
                format!("sequence of ({})", vec![
                    $(
                        $name.explain()
                    ),+
                ].join(","))
            }
        }
    }
}

mktuple!(Tuple1, 1, [A], [a]);
mktuple!(Tuple2, 2, [A, B], [a, b]);
mktuple!(Tuple3, 3, [A, B, C], [a, b, c]);
mktuple!(Tuple4, 4, [A, B, C, D], [a, b, c, d]);

#[derive(Debug)]
pub struct Stanza<T: Pattern>(pub &'static str, pub T);

impl<T: Pattern> Pattern for Stanza<T> {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        let mut r = TokenReader::new(l);

        let indent = r.expect(&Whitespace)
            .map(|m| match m {
                Captured::Vec(v) => v,
                _ => unreachable!(),
            })
            .unwrap_or(vec![]);

        let key = match r.expect(&TokenKind::Word(self.0))? {
            Captured::Token(Token {
                kind: TokenKind::Word(w),
                ..
            }) => w,
            _ => unreachable!(),
        };

        r.expect(&Whitespace)?;
        let values = Box::new(r.expect(&self.1)?);
        r.expect(&Whitespace).ok();
        r.expect(&TokenKind::Newline)?;
        r.done(Captured::Stanza {
            indent, key, values,
        })
    }

    fn explain(&self) -> String {
        format!("stanza {} containing {}", self.0, self.1.explain())
    }
}

#[derive(Debug)]
pub struct EOL;

impl Pattern for EOL {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        // Not using TokenReader as we don't have expect_not.
        let mut start: Option<usize> = None;
        let mut end: Option<usize> = None;
        let mut consumed = 0usize;
        for t in l.iter() {
            if t.kind == TokenKind::Newline {
                end = Some(t.offs);
                break
            }
            if start.is_none() {
                start = Some(t.offs);
            }
            consumed += 1;
        }

        // TODO: handle EOF. This will panic right now.
        let s = &l.parent.underlying[start.unwrap()..end.unwrap()];

        Ok(Match {
            consumed,
            captured: Captured::Str(s),
        })
    }

    fn explain(&self) -> String {
        "string until newline".to_string()
    }
}

#[derive(Debug,Clone)]
pub struct Word;

impl Pattern for Word {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        // Not using TokenReader as this is effectively (but not really) a terminal, too (ie., it's
        // TokenKind::Word but we don't have GADTs)
        if l.len() < 1 {
            return Err(MatchError::EOF {
                token_offset: l.start,
            });
        }
        let v = l.iter().next().unwrap();
        match v.kind {
            TokenKind::Word(_) => Ok(Match {
                consumed: 1,
                captured: Captured::Token(v.clone()),
            }),
            _ => Err(MatchError::PatternMismatch {
                token_offset: l.start,
                wanted: Box::new(self.clone()),
            }),
        }
    }

    fn explain(&self) -> String {
        "a word".to_string()
    }
}

pub trait GMANum: core::fmt::Debug + Clone + 'static {
    fn gma_parse(s: &str) -> Option<Captured<'_>>;
    fn explain() -> String;
}

impl GMANum for u32 {
    fn gma_parse(s: &str) -> Option<Captured<'_>> {
        s.parse::<u32>().ok().map(|v| Captured::U32(v))
    }
    fn explain() -> String {
        "number (unsigned, 32-bit)".to_string()
    }
}

#[derive(Debug, Clone)]
pub struct Num<N: GMANum>(std::marker::PhantomData<N>);

impl<N: GMANum> Pattern for Num<N> {
    fn matches<'i, 'p>(&self, l: &TokensView<'i, 'p>) -> MatchResult<'i> {
        let mut r = TokenReader::new(l);
        let n = match r.expect(&Word)? {
            Captured::Token(Token {
                kind: TokenKind::Word(w),
                ..
            }) => N::gma_parse(w),
            _ => unreachable!(),
        };
        match n {
            None => Err(MatchError::PatternMismatch {
                wanted: Box::new(self.clone()),
                token_offset: l.start,
            }),
            Some(i) => r.done(i),
        }
    }

    fn explain(&self) -> String {
        N::explain()
    }
}

impl<N: GMANum> Num<N> {
    pub fn new() -> Self {
        Self(std::marker::PhantomData)
    }
}

