use byteorder::ReadBytesExt;
use std::io::Read;

pub struct ReadError {
    pub msg: String,
    pub stack: Vec<ParseFrame>,
}

#[derive(Debug, Clone)]
pub struct ParseFrame {
    pub name: String,
    pub loc: usize,
}

impl std::fmt::Debug for ReadError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let stack = self.stack.iter().map(|pf| format!("{}@{:x}", pf.name, pf.loc)).collect::<Vec<String>>().join("->");
        f.debug_struct("ReadError")
            .field("msg", &self.msg)
            .field("stack", &stack)
            .finish()
    }
}

pub type ReadResult<T> = Result<T, ReadError>;

pub trait Serializable: Sized {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self>;
    fn read_parametrized<R: std::io::Read>(r: &mut ReadStream<R>, _params: &[u32]) -> ReadResult<Self> {
        Err(r.error("called with unexpected parameters"))
    }
}

impl<T: Serializable> Serializable for Vec<T> {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let count: u32 = r.read("count")?;
        // Safety check. A reasonable value of ~32MiB of smallest possible
        // element (u32).
        if count > ((32 << 20) / 4) {
            return Err(r.error(format!("count ({}) too large", count)));
        }
        (0..count).map(|i| r.read(format!("elem[{}]", i))).collect()
    }
    fn read_parametrized<R: std::io::Read>(r: &mut ReadStream<R>, params: &[u32]) -> ReadResult<Self> {
        let count = params[0];
        (0..count).map(|i| r.read(format!("elem[{}]", i))).collect()
    }
}

impl<const N: usize> Serializable for [f32; N] {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let mut buffer = [0f32; N];
        r.backing.read_f32_into::<byteorder::LittleEndian>(&mut buffer[..]).map_err(|e| {
            r.error(format!("{:?}", e))
        })?;
        r.pos += 4 * N;
        Ok(buffer)
    }
}

impl<T: Serializable> Serializable for Option<T> {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let count: u32 = r.read("count")?;
        if count != 0 {
            return Ok(Some(r.read("data")?));
        }
        return Ok(None);
    }
    fn read_parametrized<R: std::io::Read>(r: &mut ReadStream<R>, params: &[u32]) -> ReadResult<Self> {
        let count = params[0];
        if count != 0 {
            return Ok(Some(r.read("data")?));
        }
        return Ok(None);
    }
}

impl Serializable for u8 {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        Ok(r.bytes(1)?[0])
    }
}

impl Serializable for u32 {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let buf: [u8; 4] = r.bytes(4)?.try_into().unwrap();
        Ok(u32::from_le_bytes(buf))
    }
}

impl Serializable for i32 {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let buf: [u8; 4] = r.bytes(4)?.try_into().unwrap();
        Ok(i32::from_le_bytes(buf))
    }
}

impl Serializable for f32 {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let buf: [u8; 4] = r.bytes(4)?.try_into().unwrap();
        Ok(f32::from_le_bytes(buf))
    }
}

impl Serializable for String {
    fn read<R: std::io::Read>(r: &mut ReadStream<R>) -> ReadResult<Self> {
        let len: u32 = r.read("len")?;
        r.push("bytes");
        let bytes = r.bytes(len as usize)?;
        let bytes: Vec<u8> = bytes.into_iter().filter(|&b| b != 0).collect();
        let res = std::str::from_utf8(&bytes).map_err(|_| r.error("invalid string")).map(String::from);
        r.pop();
        res
    }
}

pub struct ReadStream<R: std::io::Read> {
    backing: std::io::BufReader<R>,
    stack: Vec<ParseFrame>,
    pos: usize,
}

impl<R: std::io::Read> ReadStream<R> {
    pub fn new(r: R) -> Self {
        Self {
            backing: std::io::BufReader::new(r),
            pos: 0,
            stack: vec![
                ParseFrame{name: "root".into(), loc: 0},
            ],
        }
    }

    fn push<S: std::string::ToString>(&mut self, s: S) {
        self.stack.push(ParseFrame {
            name: s.to_string(),
            loc: self.pos,
        });
    }

    fn pop(&mut self) {
        self.stack.pop();
    }
}

impl <R: std::io::Read> ReadStream<R> {
    pub fn bytes(&mut self, n: usize) -> ReadResult<Vec<u8>> {
        let mut buf: Vec<u8> = vec![0; n];
        self.pos += n;
        match self.backing.read_exact(&mut buf) {
            Ok(_) => return Ok(buf),
            Err(_) => return Err(self.error("eof")),
        }
    }

    pub fn tlv(&mut self) -> ReadResult<TLV> {
        let start_pos = self.pos;
        let tag: u32 = self.read("tag")?;
        let flags: u32 = self.read("flags")?;
        let length: u32 = self.read("length")?;
        Ok(TLV {
            tag, flags, length, start_pos,
        })
    }

    pub fn read<T: Serializable, S: std::string::ToString>(&mut self, ctx: S) -> ReadResult<T> {
        self.push(ctx);
        let res = T::read(self);
        self.pop();
        res
    }

    pub fn read_parametrized<T: Serializable, S: std::string::ToString>(&mut self, ctx: S, params: &[u32]) -> ReadResult<T> {
        self.push(ctx);
        let res = T::read_parametrized(self, params);
        self.pop();
        res
    }

    pub fn error<S: std::string::ToString>(&self, msg: S) -> ReadError {
        ReadError {
            msg: msg.to_string(),
            stack: self.stack.clone(),
        }
    }
}

pub struct TLV {
    pub tag: u32,
    pub flags: u32,
    pub length: u32,

    start_pos: usize,
}

impl TLV {
    pub fn check<R: std::io::Read>(&self, r: &mut ReadStream<R>) -> ReadResult<()> {
        if self.length == 0 {
            return Ok(())
        }
        let expected = self.start_pos + self.length as usize + 12usize;
        if expected != r.pos {
            return Err(ReadError {
                msg: format!("tlv length mismatch: expected end at {}, got at {}", expected, r.pos),
                stack: r.stack.clone(),
            });
        }
        Ok(())
    }
}

