use std::{io, string};

use gmfmacros::{GMISerializable, GMASerializable};
use serde::Serialize;

use crate::{
    gmi, gma,
    gma::Atom,
    gmi::Serializable as GMISerializable,
    gma::Serializable as GMASerializable,
};

#[derive(Debug,Serialize)]
pub struct GMF {
    pub version: u32,
    pub model_type: ModelType,
    pub unk1: u32,
    pub unk2: u32,

    pub scene: Scene,
    pub materials: MaterialList,
    pub objects: ObjectList,
}

impl GMF {
    pub fn read_gmi<R: io::Read>(r: R) -> gmi::ReadResult<Self> {
        let mut r = gmi::ReadStream::new(r);
        GMF::read(&mut r)
    }

    pub fn write_gma<W: io::Write>(&self, w: W) -> gma::WriteResult<()> {
        let mut w = gma::WriteStream::new(w);
        self.write("", &mut w)
    }
}

impl gmi::Serializable for GMF {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        if r.bytes(3)? != b"GMI".to_vec() {
            return Err(r.error("invalid magic"));
        }
        let version: u32 = r.read("version")?;
        if version != 3 {
            return Err(r.error(format!("unsupported version {}", version)));
        }
        let model_type = r.read("model_type")?;
        let unk1: u32 = r.read("unk1")?;
        if unk1 != 0 {
            return Err(r.error("invalid unk1"));
        }
        let unk2: u32 = r.read("unk2")?;
        if unk2 != 15 {
            return Err(r.error("invalid unk2"));
        }
        Ok(Self {
            version, model_type, unk1, unk2,
            scene: r.read("scene")?,
            materials: r.read("material_list")?,
            objects: r.read("object_list")?,
        })
    }
}

impl gma::Serializable for GMF {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit("GMA")?;
        w.emit_pair("*GABRIEL_ASCIIEXPORT", "3")?;
        self.model_type.write("*MODEL_TYPE", w)?;
        self.scene.write("", w)?;
        self.materials.write("", w)?;
        self.objects.write("", w)?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,Serialize)]
pub struct Color {
    pub r: u8,
    pub g: u8,
    pub b: u8,
    pub a: u8,
}

impl gma::Atom for Color {
    fn atom(&self) -> String {
        let mut val: u32 = 0;
        val |= (self.b as u32) << 16;
        val |= (self.g as u32) << 8;
        val |= (self.r as u32) << 0;
        // TODO: figure out alpha support, if at all.
        format!("0x{:x}", val)
    }
}


#[derive(Debug,GMISerializable,Serialize)]
pub enum MapKind {
    Ambient = 0,
    Diffuse = 1,
    Specular = 2,
    Shine = 3,
    ShineStrength = 4,
    SelfIllum = 5,
    Opacity = 6,
    FilterColor = 7,
    Bump = 8,
    Reflect = 9,
    Refract = 10,
    Displacement = 11,
    Name = 12,
    Generic = 13,
}

impl gma::Serializable for MapKind {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        match self {
            MapKind::Ambient => w.emit("*MAP_AMBIENT"),
            MapKind::Diffuse => w.emit("*MAP_DIFFUSE"),
            MapKind::Specular => w.emit("*MAP_SPECULAR"),
            MapKind::Shine => w.emit("*MAP_SHINE"),
            MapKind::ShineStrength => w.emit("*MAP_SHINESTRENGTH"),
            MapKind::SelfIllum => w.emit("*MAP_SELFILLUM"),
            MapKind::Opacity => w.emit("*MAP_OPACITY"),
            MapKind::FilterColor => w.emit("*MAP_FILTERCOLOR"),
            MapKind::Bump => w.emit("*MAP_BUMP"),
            MapKind::Reflect => w.emit("*MAP_REFLECT"),
            MapKind::Refract => w.emit("*MAP_REFRACT"),
            MapKind::Displacement => w.emit("*MAP_DISPLACEMENT"),
            MapKind::Name => w.emit("*MAP_NAME"),
            MapKind::Generic => w.emit("*MAP_GENERIC"),
        }
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum MapType {
    Explicit = 0,
    Spherical = 1,
    Cylindrical = 2,
    Shrinkwrap = 3,
    Screen = 4,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum BitmapFilter {
    Pyramidal = 0,
    SAT = 1,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum ModelType {
    #[gma_value("Basic Model")]
    BasicModel = 1,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(1,2)]
pub struct Scene {
    pub filename: String,
    pub first_frame: u32,
    pub last_frame: u32,
    pub frame_speed: u32,
    pub ticks_per_frame: u32,
    #[gma_name("SCENE_BACKGROUND_STATIC")]
    pub background: Color,
    #[gma_name("SCENE_AMBIENT_STATIC")]
    pub ambient: Color,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(7, 2)]
#[gma_name("MATERIAL_LIST")]
pub struct MaterialList {
    #[gma_name("MATERIAL")]
    pub materials: Vec<Material>,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(8, 2)]
pub struct Material {
    #[gma_name("MATERIAL_REF_NO")]
    pub ref_no: u32,
    pub name: String,
    pub class: String,
    pub ambient: Color,
    pub diffuse: Color,
    pub specular: Color,
    pub shine: f32,
    pub shine_strength: f32,
    pub transparency: f32,
    pub wiresize: f32,
    pub shading: Shading,
    #[gma_name("MATERIAL_XP_FALLOFF")]
    pub xp_falloff: f32,
    pub selfillum: f32,
    pub falloff: Falloff,
    #[gma_name("MATERIAL_XP_TYPE")]
    pub xp_type: XPType,
    pub textures: Option<TextureList>,
    pub sub: Option<MaterialList>,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum Shading {
    Other = 0,
    Blinn = 0xc,
}

#[derive(Debug, GMISerializable,Serialize)]
pub enum Falloff {
    In = 0,
    InTwoSided = 1,
}

impl gma::Serializable for Falloff {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        match self {
            Falloff::In => {
                w.emit_pair("*MATERIAL_FALLOFF", "In")?;
            },
            Falloff::InTwoSided => {
                w.emit("*MATERIAL_TWOSIDED")?;
                w.emit_pair("*MATERIAL_FALLOFF", "In")?;
            },
        }
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum XPType {
    Other = 0,
    Filter = 1,
    Additive = 3,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(14, 2)]
#[gma_name("TEXTURE_LIST")]
pub struct TextureList {
    #[gma_name("TEXTURE")]
    pub textures: Vec<Texture>,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(15, 4)]
pub struct Texture {
    #[gma_name("MAP_NAME")]
    pub name: String,
    #[gma_name("MAP_CLASS")]
    pub class: String,
    #[gma_name("BITMAP")]
    pub bitmap: String,
    #[gma_name("MAP_AMOUNT")]
    pub amount: f32,
    pub kind: MapKind,
    #[gma_name("MAP_TYPE")]
    pub map_type: MapType,
    #[gma_name("UVW_U_OFFSET")]
    pub u_offset: f32,
    #[gma_name("UVW_V_OFFSET")]
    pub v_offset: f32,
    #[gma_name("UVW_U_TILING")]
    pub u_tiling: f32,
    #[gma_name("UVW_V_TILING")]
    pub v_tiling: f32,
    #[gma_name("UVW_ANGLE")]
    pub angle: f32,
    #[gma_name("UVW_BLUR")]
    pub blur: f32,
    #[gma_name("UVW_BLUR_OFFSET")]
    pub blur_offset: f32,
    #[gma_name("UVW_NOUSE_AMT")]
    pub noise_amount: f32,
    #[gma_name("UVW_NOISE_SIZE")]
    pub noise_size: f32,
    #[gma_name("UVW_NOISE_LEVEL")]
    pub noise_level: u32,
    #[gma_name("UVW_NOISE_PHASE")]
    pub noise_phase: f32,
    #[gma_name("")]
    pub invert: u32,
    #[gma_name("")]
    pub unknown: u32,
    #[gma_name("BITMAP_FILTER")]
    pub filter: BitmapFilter,
    #[gma_name("BITMAP_MAP_CHANNEL")]
    pub channel: u32,
    pub sub: Option<TextureList>,
}

#[derive(Debug,Serialize)]
pub struct ObjectName(String);

impl gmi::Serializable for ObjectName {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        String::read(r).map(|e| ObjectName(e))
    }
}

impl gma::Serializable for ObjectName {
    fn write<W: io::Write, S: string::ToString>(&self, name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        let val = match self.0.len() {
            0 => "(null)".to_string(),
            _ => self.0.clone(),
        };
        val.write(name, w)
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged_nolen(18,2)]
#[gma_name("OBJECT_LIST")]
pub struct ObjectList {
    #[gma_name("OBJECT")]
    pub objects: Vec<Object>,
}

#[derive(Debug,Serialize)]
pub enum Object {
    Geometry(GeometryObject),
    Camera(CameraObject),
    Light(LightObject),
    AttachmentPoint(AttachmentPointObject),
    Simulation(SimulationObject),
    RBCollection(RBCollectionObject),
    ConstraintSolver(ConstraintSolverObject),
    AngularDashpot(AngularDashpotObject),
}

impl gmi::Serializable for Object {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        let tlv = r.tlv()?;
        let (res, check) = match (tlv.tag, tlv.flags) {
            (2, 4) => (Object::Geometry(r.read("geometry")?), false),
            (4, 2) => (Object::Camera(r.read("camera")?), false),
            (5, 3) => (Object::Light(r.read("light")?), false),
            (21, 2) => (Object::AttachmentPoint(r.read("attachmentpoint")?), true),
            (30, 2) => (Object::Simulation(r.read("simulation")?), false),
            (31, 4) => (Object::RBCollection(r.read("rigidbodycollection")?), true),
            (42, 3) => (Object::ConstraintSolver(r.read("constraintsolver")?), true),
            (49, 2) => (Object::AngularDashpot(r.read("angulardashpot")?), true),
            _ => return Err(r.error(format!("unknown object type ({}, {})", tlv.tag, tlv.flags))),
        };
        if check {
            tlv.check(r)?;
        }
        Ok(res)
    }
}

impl gma::Serializable for Object {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        match self {
            Object::Geometry(o) => o.write("", w),
            Object::Camera(o) => o.write("", w),
            Object::Light(o) => o.write("", w),
            Object::AttachmentPoint(o) => o.write("", w),
            Object::Simulation(o) => o.write("", w),
            Object::RBCollection(o) => o.write("", w),
            Object::ConstraintSolver(o) => o.write("", w),
            Object::AngularDashpot(o) => o.write("", w),
        }
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GEOMOBJECT")]
pub struct GeometryObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    #[gma_name("NODE_PARENT")]
    pub parent: String,
    #[gma_name("NODE_SHADEVERTS")]
    pub shade_verts: u8,
    #[gma_name("NODE_TM")]
    pub tm: TransformMatrix,
    pub mesh: Mesh,
}

#[derive(Debug,GMISerializable,Serialize)]
pub struct CameraObject {
    pub name: String,
    pub tm1: TransformMatrix,
    pub tm2: Option<TransformMatrix>,
    pub type_: CameraType,
    pub hither: f32,
    pub yon: f32,
    pub near: f32,
    pub far: f32,
    pub fov: f32,
    pub tdist: f32,
}

impl gma::Serializable for CameraObject {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {

        w.push("*CAMERA")?;
        self.name.write("*NODE_NAME", w)?;
        self.type_.write("*CAMERA_TYPE", w)?;
        self.tm1.write("*NODE_TM", w)?;
        self.tm2.write("*NODE_TM", w)?;
        self.hither.write("*CAMERA_HITHER", w)?;
        self.yon.write("*CAMERA_YON", w)?;
        self.near.write("*CAMERA_NEAR", w)?;
        self.far.write("*CAMERA_FAR", w)?;
        self.fov.write("*CAMERA_FOV", w)?;
        self.tdist.write("*CAMERA_TDIST", w)?;
        w.pop()?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum CameraType {
    Target = 0,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged(17, 2)]
#[gma_name("NODE_TM")]
pub struct TransformMatrix {
    #[gma_name("NODE_NAME")]
    pub name: String,
    #[gma_name("TM_ROW0")]
    pub row0: [f32; 3],
    #[gma_name("TM_ROW1")]
    pub row1: [f32; 3],
    #[gma_name("TM_ROW2")]
    pub row2: [f32; 3],
    #[gma_name("TM_ROW3")]
    pub row3: [f32; 3],
}

#[derive(Debug, GMISerializable,Serialize)]
#[gmi_tagged_nolen(16, 4)]
pub struct Mesh {
    pub time: u32,
    pub vertex_count: u32,
    pub face_count: u32,
    pub tvertex_count: u32,
    pub cvertex_count: u32,
    pub material_ref: i32,

    #[gmi_read_parameters(vertex_count)]
    pub vertices: MeshVertexList,
    #[gmi_read_parameters(face_count)]
    pub faces: MeshFaceList,

    #[gmi_read_parameters(tvertex_count)]
    pub tvertices: MeshTVertexList,
    #[gmi_read_parameters(if tvertex_count > 0 { face_count } else { 0 })]
    pub tfaces: MeshTFaceList,

    pub channels: Option<Vec<TextureChannel>>,

    #[gmi_read_parameters(cvertex_count)]
    pub cvertices: MeshCVertexList,
    #[gmi_read_parameters(if cvertex_count > 0 { face_count } else { 0 })]
    pub cfaces: MeshCFaceList,

    #[gmi_read_parameters(face_count)]
    pub normals: MeshNormalList,

    pub backface_cull: u32,
}

impl gma::Serializable for Mesh {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.push("*MESH")?;
        self.time.write("*TIMEVALUE", w)?;
        self.vertex_count.write("*MESH_NUMVERTEX", w)?;
        self.face_count.write("*MESH_NUMFACES", w)?;
        self.vertices.write("", w)?;
        self.faces.write("", w)?;
        if self.tvertex_count > 0 {
            self.tvertex_count.write("*MESH_NUMTVERTEX", w)?;
            self.tvertices.write("", w)?;
            self.face_count.write("*MESH_NUMTVFACES", w)?;
            self.tfaces.write("", w)?;
        }
        if let Some(channels) = &self.channels {
            for (i, channel) in channels.iter().enumerate() {
                w.push(format!("*MESH_MAPPINGCHANNEL\t{}", i+2))?;
                channel.tvertex_count.write("*MESH_NUMTVERTEX", w)?;
                channel.tvertices.write("", w)?;
                channel.face_count.write("*MESH_NUMTVFACES", w)?;
                channel.tfaces.write("", w)?;
                w.pop()?;
            }
        }
        if self.cvertices.0.len() > 0 {
            self.cvertices.0.len().write("*MESH_NUMCVERTEX", w)?;
            self.cvertices.write("", w)?;
        }
        if self.cfaces.0.len() > 0 {
            self.cfaces.0.len().write("*MESH_NUMCVFACES", w)?;
            self.cfaces.write("", w)?;
        }
        if self.normals.0.len() > 0 {
            self.normals.write(w, &self.faces)?;
        }
        self.backface_cull.write("*BACKFACE_CULL", w)?;
        self.material_ref.write("*MATERIAL_REF", w)?;
        w.pop()?;
        Ok(())
    }
}

macro_rules! specialized_vec {
    ($newname:ident, $elemname:ident, $header:expr, $format:expr) => {
        #[derive(Debug,Serialize)]
        pub struct $newname(Vec<$elemname>);
        impl gmi::Serializable for $newname {
            fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
                <Vec<$elemname> as gmi::Serializable>::read(r).map(|r| $newname(r))
            }
            fn read_parametrized<R: io::Read>(r: &mut gmi::ReadStream<R>, params: &[u32]) -> gmi::ReadResult<Self> {
                <Vec<$elemname> as gmi::Serializable>::read_parametrized(r, params).map(|r| $newname(r))
            }
        }
        impl gma::Serializable for $newname {
            fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
                w.push($header)?;
                for (i, v) in self.0.iter().enumerate() {
                    w.emit(&$format(i, v))?;
                }
                w.pop()?;
                Ok(())
            }
        }

    }
}

specialized_vec!(
    MeshVertexList, Point,
    "*MESH_VERTEX_LIST",
    (|i, v: &Point| {
        format!("*MESH_VERTEX\t{}\t{:.6}\t{:.6}\t{:.6}", i, v.x, v.y, v.z)
    })
);

specialized_vec!(
    MeshFaceList, Face,
    "*MESH_FACE_LIST",
    (|i, f: &Face| {
        format!(
            "*MESH_FACE\t{:>4}\tA:{:>5}\tB:{:>5}\tC:{:>5}\t*MESH_MTLID {}",
            i, f.a, f.b, f.c, f.mtlid)
    })
);

specialized_vec!(
    MeshTVertexList, Point,
    "*MESH_TVERTLIST",
    (|i, v: &Point| {
        format!("*MESH_TVERT\t{}\t{:.6}\t{:.6}\t{:.6}", i, v.x, v.y, v.z)
    })
);

specialized_vec!(
    MeshTFaceList, TFace,
    "*MESH_TFACELIST",
    (|i, f: &TFace| {
        format!("*MESH_TFACE\t{}\t{}\t{}\t{}", i, f.a, f.b, f.c)
    })
);

specialized_vec!(
    MeshCVertexList, Point,
    "*MESH_CVERTLIST",
    (|i, v: &Point| {
        format!("*MESH_VERTCOL\t{}\t{:.6}\t{:.6}\t{:.6}", i, v.x, v.y, v.z)
    })
);

specialized_vec!(
    MeshCFaceList, TFace,
    "*MESH_CFACELIST",
    (|i, f: &TFace| {
        format!("*MESH_CFACE\t{}\t{}\t{}\t{}", i, f.a, f.b, f.c)
    })
);


specialized_vec!(
    MeshNormalList, FaceNormal,
    "*MESH_NORMALS",
    (|_, _: &FaceNormal| -> String {
        // This implementation is never used.
        unreachable!()
    })
);

impl MeshNormalList {
    fn write<W: io::Write>(&self, w: &mut gma::WriteStream<W>, faces: &MeshFaceList) -> gma::WriteResult<()> {
        w.push("*MESH_NORMALS")?;
        for (i, (normal, face)) in self.0.iter().zip(faces.0.iter()).enumerate() {
            w.emit(&format!("*MESH_FACENORMAL\t{}\t{:.6}\t{:.6}\t{:.6}",
                            i, normal.face[0], normal.face[1], normal.face[2]))?;
            for j in 0..3 {
                let vnx = normal.vertex[j];
                let vny = normal.vertex[j+1];
                let vnz = normal.vertex[j+2];
                let vno = match j {
                    0 => face.a,
                    1 => face.b,
                    2 => face.c,
                    _ => unreachable!(),
                };
                w.emit(&format!("\t*MESH_VERTEXNORMAL\t{}\t{:.6}\t{:.6}\t{:.6}",
                                vno, vnx, vny, vnz))?;
            }
        }
        w.pop()?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub struct Point {
    pub x: f32,
    pub y: f32,
    pub z: f32,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub struct Face {
    pub a: u32,
    pub b: u32,
    pub c: u32,
    pub mtlid: u32,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub struct TFace {
    pub a: u32,
    pub b: u32,
    pub c: u32,
}

#[derive(Debug,GMISerializable,Serialize)]
pub struct FaceNormal {
    pub face: [f32; 3],
    pub vertex: [f32; 9],
}

impl gma::Serializable for FaceNormal {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, _w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        Ok(())
    }
}

#[derive(Debug, GMISerializable,GMASerializable,Serialize)]
pub struct TextureChannel {
    pub unk1: u32,
    pub tvertex_count: u32,
    pub unk2: u32,
    pub face_count: u32,
    pub unk3: u32,
    pub unk4: u32,
    pub tvertex_count2: u32,
    pub face_count2: u32,

    #[gmi_read_parameters(tvertex_count)]
    pub tvertices: MeshTVertexList,
    #[gmi_read_parameters(face_count)]
    pub tfaces: MeshTFaceList,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("LIGHT")]
pub struct LightObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    pub tm2: Option<TransformMatrix>,
    #[gma_name("LIGHT_TYPE")]
    pub light_type: LightType,
    pub shadows: LightShadows,
    pub uselight: u32,
    pub spotshape: LightSpotShape,
    pub color: Color,
    #[gma_name("LIGHT_INTENS")]
    pub intensity: f32,
    pub aspect: f32,
    #[gma_skip(self.light_type == LightType::Omni)]
    pub hotspot: f32,
    #[gma_skip(self.light_type == LightType::Omni)]
    pub falloff: f32,
    #[gma_skip(self.light_type == LightType::Directional)]
    pub attn_start: f32,
    #[gma_skip(self.light_type == LightType::Directional)]
    pub attn_end: f32,
    pub tdist: f32,
    pub use_far_attenuation: FarAttenuation,
}

#[derive(Debug,PartialEq,Eq,GMISerializable,GMASerializable,Serialize)]
pub enum LightType {
    Omni = 0,
    Target = 1,
    Directional = 2,
    Free = 4,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum LightShadows {
    Off = 0,
    Mapped = 1,
    Raytraced = 2,
}

#[derive(Debug,GMASerializable,Serialize)]
pub enum LightSpotShape {
    Circle = 0,
}

impl gmi::Serializable for LightSpotShape {
    fn read<R: io::Read>(_: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        Ok(LightSpotShape::Circle)
    }
}

#[derive(Debug,Serialize)]
pub struct FarAttenuation(u32);

impl gmi::Serializable for FarAttenuation {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        // Uninitialized memory in stock .GMFs. Whoops.
        let v = u32::read(r)? & 0xff;
        Ok(FarAttenuation(v))
    }
}

impl gma::Serializable for FarAttenuation {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit(&format!("*USE FAR ATTENUATION = \t{}", self.0))
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_ATTACHMENTPT")]
pub struct AttachmentPointObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    #[gma_name_bare("USER DATA")]
    pub user_data: String,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_CONSTRAINTSOLVER")]
pub struct ConstraintSolverObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    #[gma_name("THRESHOLD")]
    pub threshold: f32,
    #[gma_name("RB_COLLECTION_NAME")]
    pub rb_collection_name: String,
    pub constraints: Option<ConstraintList>,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged_nolen(44, 2)]
#[gma_name("GMID_HAVOK_CONSTRAINT_LIST")]
pub struct ConstraintList {
    #[gma_name("")]
    pub constraints: Vec<Constraint>,
}

#[derive(Debug,Serialize)]
pub enum Constraint {
    Wheel(WheelConstraint),
    Hinge(HingeConstraint),
    PointToPoint(PointToPointConstraint),
}

impl gmi::Serializable for Constraint {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        let tlv = r.tlv()?;
        let res = match (tlv.tag, tlv.flags) {
            (46, 2) => Constraint::Wheel(r.read("pivot")?),
            (47, 2) => Constraint::Hinge(r.read("hinge")?),
            (53, 2) => Constraint::PointToPoint(r.read("pointtopoint")?),
            _ => return Err(r.error(format!("unknown constraint type ({}, {})", tlv.tag, tlv.flags))),
        };
        tlv.check(r)?;
        Ok(res)
    }
}

impl gma::Serializable for Constraint {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        match self {
            Constraint::Wheel(o) => o.write("", w),
            Constraint::Hinge(o) => o.write("", w),
            Constraint::PointToPoint(o) => o.write("", w),
        }
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_WHEEL_CONSTRAINT")]
pub struct WheelConstraint {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    #[gma_name("BODY1")]
    pub body1: String,
    #[gma_name("BODY2")]
    pub body2: String,
    #[gma_name("POINT")]
    pub point: [f32; 3],
    #[gma_name("SPIN_AXIS")]
    pub spin_axis: [f32; 3],
    #[gma_name("SUSPENSION_AXIS")]
    pub suspensions_axis: [f32; 3],
    #[gma_name("SUSPENSION_LIMITS")]
    pub suspensions_limits: [f32; 2],
    #[gma_name("SUSPENSION_FRICTION")]
    pub suspension_fiction: f32,
    #[gma_name("ANGULAR_SPEED")]
    pub angular_speed: f32,
    #[gma_name("GAIN")]
    pub gain: f32,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_HINGE_CONSTRAINT")]
pub struct HingeConstraint {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    #[gma_name("BODY1")]
    pub body1: String,
    #[gma_name("BODY2")]
    pub body2: String,
    #[gma_name("POINT")]
    pub point: [f32; 3],
    #[gma_name("SPIN_AXIS")]
    pub spin_axis: [f32; 3],
    #[gma_name("IS_LIMITED")]
    #[gma_space_delim]
    pub is_limited: u8,
    #[gma_name("FRICTION")]
    #[gma_space_delim]
    pub friction: f32,
    #[gma_name("ANGLE_LIMITS")]
    pub angle_limits: [f32; 2],
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_POINTTOPOINT")]
pub struct PointToPointConstraint {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    #[gma_name("BODY1")]
    pub body1: String,
    #[gma_name("BODY2")]
    pub body2: String,
    pub points: PTPPoints,
}

#[derive(Debug,GMISerializable,Serialize)]
pub struct PTPPoints {
    a: Point,
    b: Point,
}

impl gma::Serializable for PTPPoints {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit("*POINT 1")?;
        w.emit(&format!("(  {:.6}\t{:.6}\t{:.6} )", self.a.x, self.a.y, self.a.z))?;
        w.emit(&format!("(  {:.6}\t{:.6}\t{:.6} )", self.b.x, self.b.y, self.b.z))?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_SIMOBJECT")]
pub struct SimulationObject {
    #[gma_name("NODE_NAME")]
    pub name: ObjectName,
    #[gma_name("GRAVITY")]
    pub gravity: Gravity,
    #[gma_name("WORLDSCALE")]
    pub worldscale: f32,
    #[gma_name("SIMTOLERANCE")]
    pub simtolerance: f32,
    #[gma_name("RESOLVER")]
    pub resolver: u32,
    #[gma_name("INCLUDE_DRAG")]
    pub incl_drag: u8,
    #[gma_name("LINEAR_DRAG")]
    pub linear_drag: f32,
    #[gma_name("ANGULAR_DRAG")]
    pub angular_drag: f32,
    #[gma_name("INCLUDE_DEACTIVATOR")]
    pub incl_deactivator: u8,
    #[gma_name("SHORTFREQ")]
    pub shortfreq: f32,
    #[gma_name("LONGFREQ")]
    pub longfreq: f32,
    #[gma_name("USE_FAST_SUBSPACE")]
    pub use_fast_subspace: u8,
    #[gma_name("UPDATES_PER_TIMESTEP")]
    pub updates_per_timestamp: f32,
    #[gma_name("NUM_COLLISION_PAIRS")]
    pub collision_pairs: u32,
}

#[derive(Debug,Serialize)]
pub struct Gravity(Point);

impl gmi::Serializable for Gravity {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        Point::read(r).map(|e| Gravity(e))
    }
}

impl gma::Serializable for Gravity {
    fn write<W: io::Write, S: string::ToString>(&self, name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit(&format!("{}\t{:.6} {:.6} {:.6}",
                        name.to_string(), self.0.x, self.0.y, self.0.z))
    }
}


#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_RBCOLLECTION")]
pub struct RBCollectionObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    #[gma_name("NUM_DISABLED_PAIRS")]
    pub disabled_pairs: u32,
    #[gma_name("SOLVER_TYPE")]
    pub solver_type: u32,
    pub rigidbody_list: RigidBodyList,
    // Some DSL-IRL 2.6 GMFs have this set to ffffff / ffffffff for some reason..
    #[gmi_read_parameters(if disabled_pairs < 4096 { disabled_pairs } else { 0 })]
    pub disabled_collision_pairs: Option<DisabledCollisionPairList>,
}

#[derive(Debug,GMISerializable,Serialize)]
#[gmi_tagged_nolen(33, 2)]
pub struct RigidBodyList {
    pub rigidbodies: Vec<RigidBody>,
}

impl gma::Serializable for RigidBodyList {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        self.rigidbodies.len().write("*COUNT", w)?;
        w.push("*GMID_HAVOK_RIGIDBODY_LIST")?;
        self.rigidbodies.len().write("*COUNT", w)?;
        for rigidbody in self.rigidbodies.iter() {
            rigidbody.write("", w)?;
        }
        w.pop()?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged_nolen(32, 4)]
#[gma_name("GMID_HAVOK_RIGIDBODY")]
pub struct RigidBody {
    #[gma_name("NODE_NAME")]
    pub name: String,
    #[gma_name("MASS")]
    pub mass: f32,
    #[gma_name("ELASTICITY")]
    pub elasticity: f32,
    #[gma_name("FRICTION")]
    pub friction: f32,
    #[gma_name("OPTIMIZATION_LEVEL")]
    pub optimization_level: f32,
    #[gma_name("UNYIELDING")]
    pub unyielding: u32,
    #[gma_name("SIMULATION_GEOMETRY")]
    pub simulation_geometry: u32,
    #[gma_name("GEOMETRY_PROXY_NAME")]
    pub geometry_proxy_name: ObjectName,
    #[gma_name("USE_DISPLAY_PROXY")]
    pub use_display_proxy: u8,
    #[gma_name("DISABLE_COLLISIONS")]
    pub disable_collisions: u8,
    #[gma_name("INACTIVE")]
    pub inactive: u8,
    #[gma_name("DISPLAY_PROXY_NAME")]
    pub display_proxy_name: ObjectName,
    pub tm: TransformMatrix,
    #[gma_name("HAVOK_GEO_TYPE")]
    pub geo_type: GeoType,
    pub children: NestedRigidBodyList,
}

#[derive(Debug,Serialize)]
pub struct NestedRigidBodyList(Option<RigidBodyList>);

impl gmi::Serializable for NestedRigidBodyList {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        Option::<RigidBodyList>::read(r).map(|el| NestedRigidBodyList(el))
    }
}

impl gma::Serializable for NestedRigidBodyList {
    fn write<W: io::Write, S: string::ToString>(&self, name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        let count = self.0.as_ref().map(|l| l.rigidbodies.len()).unwrap_or(0);
        count.write("*NUMBER_OF_CHILDREN", w)?;
        self.0.write(name, w)?;
        Ok(())
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
pub enum GeoType {
    Standard = 0,
    Plane = 1,
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gmi_tagged_nolen(51, 2)]
#[gma_name("GMID_HAVOK_DIS_COLLISION_PAIRS")]
pub struct DisabledCollisionPairList {
    #[gma_name("")]
    pub pairs: Vec<DisabledCollisionPair>,
}

#[derive(Debug,GMISerializable,Serialize)]
pub struct DisabledCollisionPair {
    pub a: String,
    pub b: String,
}

impl gma::Serializable for DisabledCollisionPair {
    fn write<W: io::Write, S: string::ToString>(&self, _name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit(&format!("{{ {}\t{}  }}",
                        self.a, self.b))
    }
}

#[derive(Debug,GMISerializable,GMASerializable,Serialize)]
#[gma_name("GMID_HAVOK_ANGULAR_DASHPOT")]
pub struct AngularDashpotObject {
    #[gma_name("NODE_NAME")]
    pub name: String,
    pub tm: TransformMatrix,
    #[gma_name("BODY1")]
    pub body1: String,
    #[gma_name("BODY2")]
    pub body2: String,
    #[gma_name("STRENGTH")]
    pub strength: f32,
    #[gma_name("DAMPING")]
    pub damping: f32,
    #[gma_name("ALLOW_INTERPENETRATIONS")]
    pub allow_interpenetrations: u8,
    #[gma_name("QUATERNION")]
    pub quaternion: Quaternion,
}

#[derive(Debug,Serialize)]
pub struct Quaternion([f32; 4]);

impl gmi::Serializable for Quaternion {
    fn read<R: io::Read>(r: &mut gmi::ReadStream<R>) -> gmi::ReadResult<Self> {
        <[f32; 4]>::read(r).map(|el| Quaternion(el))
    }
}

impl gma::Serializable for Quaternion {
    fn write<W: io::Write, S: string::ToString>(&self, name: S, w: &mut gma::WriteStream<W>) -> gma::WriteResult<()> {
        w.emit(&name.to_string())?;
        let values = self.0.iter().map(|v| format!("{:.6}", v)).collect::<Vec<_>>().join("\t");
        w.emit(&format!("(  {} )", values))?;
        Ok(())
    }
}

